SystemOrganization addCategory: #NumberPrinter!

Object subclass: #FloatPrinter
	instanceVariableNames: 'base r s mPlus mMinus roundingLowIncludesLimits roundingHighIncludesLimits'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'NumberPrinter'!

!FloatPrinter commentStamp: 'nice 6/10/2012 18:34' prior: 0!
A FloatPrinter is class specialized in printing floating point numbers
The minimum number of digits necessary for reconstructing the exact value of the Float is used.
Beside this minimum, a trailing fraction separator and zero can be appended as fraction part to make sure that the number is reinterpreted asFloat.

Default behaviour is to print a Float in free format.
Free format uses either scientific format or fixed format, depending on the number scale.
The rules are
- no more than 6 digits left of fraction separator if magnitude is greater than or equal to one,
- and no more than 3 leading fractional zeros if magnitude is smaller than one.

Formatting variations should be implemented in subclasses.
This class responsibility is rather to provide the basic arithmetic involved in printing accurately.
It does it based upon the algorithm outlined in:
	Robert G. Burger and R. Kent Dybvig
	Printing Floating Point Numbers Quickly and Accurately
	ACM SIGPLAN 1996 Conference on Programming Language Design and Implementation
	June 1996.
This algorithm is itself a variation of Guy Steele's dragon 4 implemented originally in scheme.

Main internal variables of the algorithm are stored in instance variables so as to reduce the number of in/out parameters versus a pure functional implementation. The private methods contains some hints about the algorithm in comments, along with this class comment, but it is recommanded to read the paper for a better understanding.

Instance Variables
	base:		<Integer>
	mMinus:		<Integer | Fraction>
	mPlus:		<Integer | Fraction>
	r:		<Integer>
	roundingHighIncludesLimits:		<Boolean>
	roundingLowIncludesLimits:		<Boolean>
	s:		<Integer>

base
	- the base into which the number shall be printed, usually 10

mMinus
	- half of difference (aFloat abs - aFloat abs predecessor) scaled by s

mPlus
	- half of difference (aFloat abs successor - aFloat abs) scaled by s

r
	- the mantissa (scaled by s)

roundingHighIncludesLimits
	- tells whether (r+mPlus)/s is within the limits or not (that is, represents the same float)

roundingLowIncludesLimits
	- tells whether (r-mMinus)/s is within the limits or not (that is, represents the same float)

s
	- the scale used to have integer representation rather than fractional

Note that the variables r, s, mPlus and mMinus are changing during algorithm progress.
r/s is initially the Float to be printed.
r or s is then scaled by appropriate power of base such that 1 <= (r/s) < base.
r and s are then modified by quotient/remainder divisions in the process of generating the digits.

These states wer removed and are now handled as a temporary in various places:
baseExpEstimate
	- an estimate of aFloat exponent in printed base, that is
	(base raisedTo: baseExpEstimate - 1) <= aFloat < (base raisedTo: baseExpEstimate)
	
decPointCount
	- the number of digits remaining before decimal point!

!FloatPrinter class methodsFor: 'factory' stamp: 'nice 6/1/2012 23:06'!
basic
	^FloatPrinter new! !

!FloatPrinter class methodsFor: 'factory' stamp: 'nice 6/1/2012 22:21'!
fixed
	^FloatPrinterFixedFormat new! !

!FloatPrinter class methodsFor: 'factory' stamp: 'nice 6/1/2012 22:21'!
forStore
	^FloatPrinterForStore new! !

!FloatPrinter class methodsFor: 'factory' stamp: 'nice 6/2/2012 17:44'!
freeFormat
	^FloatPrinterFreeFormat new! !

!FloatPrinter class methodsFor: 'factory' stamp: 'nice 9/27/2021 14:56'!
hex
	^FloatPrinterHexadecimalFormat new! !

!FloatPrinter class methodsFor: 'factory' stamp: 'nice 6/1/2012 22:22'!
scientific
	^FloatPrinterScientificFormat new! !

!FloatPrinter class methodsFor: 'factory' stamp: 'nice 4/7/2016 21:21'!
thousand
	^FloatPrinterThousandFormat new! !

!FloatPrinter methodsFor: 'private' stamp: 'nice 6/8/2012 21:31'!
adjustEstimatedBaseExponent: baseExpEstimate forRoundingToDigitCount: nDigits
	"Adjust the estimate of exponent in target printed base when restricting the number of printed digits to prescribed count cause a rounding to an upper power of base.
	Since 1 <= (r/s) < base on input, such rounding occurs if (r/s) + ((base/2)*(1/(base raisedTo: nDigits))) >= base"
	
	| bs |
	(bs := base * s) - r * (base raisedToInteger: nDigits) > (bs / 2)
		ifFalse:
			[baseExpEstimate >= 0
				ifTrue: [s := bs]
				ifFalse:
					[r := r / base.
					mPlus := mPlus / base.
					mMinus := mMinus / base].
			^baseExpEstimate + 1].
	^baseExpEstimate! !

!FloatPrinter methodsFor: 'accessing' stamp: 'nice 6/1/2012 23:58'!
base: anInteger
	"Set the target base into which the float must be printed"
	base := anInteger! !

!FloatPrinter methodsFor: 'accessing' stamp: 'nice 6/1/2012 23:19'!
digitCount: anIntegerOrNil
	anIntegerOrNil ifNotNil: [self error: 'digitCount cannot be specified in a basic FloatPrinter, use a versatile one']! !

!FloatPrinter methodsFor: 'private' stamp: 'nice 6/8/2012 21:36'!
estimateBaseExponent: base2Exponent
	"Estimate the exponent in target printed base, and scale the integer representation of the float to be printed by according power.
	This exponent is adjusted to take into account the interval represented by the Float, because we always want the shortest possible representation.
	On output, scaling is applied with appropriate power of base, such that 1 <= (r/s) < base"
	| scale baseExpEstimate |
	baseExpEstimate := (base2Exponent * base asFloat reciprocalLogBase2 - 1.0e-10) ceiling.
	baseExpEstimate >= 0
		ifTrue: [s := s * (base raisedToInteger: baseExpEstimate)]
		ifFalse:
			[scale := base raisedToInteger: baseExpEstimate negated.
			r := r * scale.
			mPlus := mPlus * scale.
			mMinus := mMinus * scale].
	^((r + mPlus >= s) and: [roundingHighIncludesLimits or: [r + mPlus > s]])
		ifTrue: [baseExpEstimate + 1]
		ifFalse:
			[r := r * base.
			mPlus := mPlus * base.
			mMinus := mMinus * base.
			baseExpEstimate].! !

!FloatPrinter methodsFor: 'accessing' stamp: 'nice 6/1/2012 23:19'!
exponentChar: aCharacter
	aCharacter = $e ifFalse: [self error: 'exponentChar cannot be specified for a basic FloatPrinter, use a versatile one']! !

!FloatPrinter methodsFor: 'accessing' stamp: 'nice 6/1/2012 23:20'!
forceSign: aBoolean
	aBoolean ifTrue: [self error: 'plus sign cannot be forced for a basic FloatPrinter, use a versatile one']! !

!FloatPrinter methodsFor: 'accessing' stamp: 'nice 6/1/2012 22:24'!
fractionSeparator
	^$.! !

!FloatPrinter methodsFor: 'accessing' stamp: 'nice 6/1/2012 23:20'!
fractionSeparator: aCharacter
	aCharacter = $. ifFalse: [self error: 'fractionSeparator cannot be specified for a basic FloatPrinter, use a versatile one']! !

!FloatPrinter methodsFor: 'private' stamp: 'nice 6/7/2012 23:13'!
generateDigitsOn: aStream decPointCount: anInteger
	"Print a serie of digits representing the float to prescribed precision..
	This code will eventually print the fractionSeparator if it lies in between those digits, but won't print the exponent.
	Note that the two terminating conditions tc1 and tc2 are here for detecting that printed representation lie in the interval representing the float to be printed.
	Note also the eventual adjustment of last digit to account for correct rounding."
	| d shead slowbit tc1 tc2 decPointCount |
	self simplifyFraction.
	mPlus = mMinus ifTrue: [^self generateDigitsSymOn: aStream decPointCount: anInteger].
	decPointCount := anInteger.
	slowbit := 1 - s lowBit .
	shead := s bitShift: slowbit.
	[d := (r bitShift: slowbit) // shead.
	r := r - (d * s).
	(tc1 := (r <= mMinus) and: [roundingLowIncludesLimits or: [r < mMinus]]) |
	(tc2 := (r + mPlus >= s) and: [roundingHighIncludesLimits or: [r + mPlus > s]])] whileFalse:
		[aStream nextPut: (Character digitValue: d).
		r := r * base.
		mPlus := mPlus * base.
		mMinus := mMinus * base.
		(decPointCount := decPointCount - 1) = 0 ifTrue: [aStream nextPut: self fractionSeparator]].
	tc2 ifTrue:
		[(tc1 not or: [r * 2 >= s]) ifTrue: [d := d + 1]].
	aStream nextPut: (Character digitValue: d).
	^decPointCount! !

!FloatPrinter methodsFor: 'private' stamp: 'nice 6/7/2012 23:12'!
generateDigitsSymOn: aStream decPointCount: anInteger
	"Special case of generateDigitsOn: when mPlus = mMinus.
	Since this is most frequent case and greater CPU contributor, this optimization is worth."
	| d shead slowbit s1 tc1 tc2 decPointCount |
	decPointCount := anInteger.
	slowbit := 1 - s lowBit .
	shead := s bitShift: slowbit.
	s1 := roundingHighIncludesLimits ifTrue: [s - 1] ifFalse: [s].
	[d := (r bitShift: slowbit) // shead.
	r := r - (d * s).
	(tc1 := (r <= mPlus) and: [roundingLowIncludesLimits or: [r < mPlus]]) |
	(tc2 := s1 - r < mPlus)] whileFalse:
		[aStream nextPut: (Character digitValue: d).
		r := r * base.
		mPlus := mPlus * base.
		(decPointCount := decPointCount - 1) = 0 ifTrue: [aStream nextPut: self fractionSeparator]].
	tc2 ifTrue:
		[(tc1 not or: [r * 2 >= s]) ifTrue: [d := d + 1]].
	aStream nextPut: (Character digitValue: d).
	^decPointCount! !

!FloatPrinter methodsFor: 'initialize-release' stamp: 'nice 6/1/2012 22:20'!
initialize
	base := 10! !

!FloatPrinter methodsFor: 'printing' stamp: 'nice 6/2/2012 20:10'!
print: aFloat
	^String new: 32 streamContents: [:aStream | self print: aFloat on: aStream]! !

!FloatPrinter methodsFor: 'printing' stamp: 'nice 6/10/2012 21:36'!
print: aNumber on: aStream
	"Print aNumber on aStream in free format.
	The printed representation will have the minimum number of digits so that aNumber can be reconstructed exactly."
	| fixedFormat baseExpEstimate decPointCount |
	aNumber isFinite ifFalse: [^self printExceptionalFloat: aNumber on: aStream].
	self printSignOf: aNumber on: aStream.
	aNumber = 0.0 ifTrue: [aStream nextPutAll: '0.0'. ^ self].
	baseExpEstimate := aNumber initializePrinter: self.
	(fixedFormat := baseExpEstimate between: -3 and: 6)
		ifTrue:
			[(decPointCount := baseExpEstimate) <= 0
				ifTrue: [aStream nextPutAll: ('0.000' truncateTo: 2 - baseExpEstimate)]]
		ifFalse:
			[decPointCount := 1].
	decPointCount := self generateDigitsOn: aStream decPointCount: decPointCount.
	decPointCount > 0
		ifTrue:
			[aStream nextPutAll: ('00000.0' last: decPointCount +1)].
	fixedFormat
		ifFalse:
			[aStream nextPut: $e.
			baseExpEstimate - 1 printOn: aStream]! !

!FloatPrinter methodsFor: 'private-printing' stamp: 'nice 6/1/2012 22:27'!
printExceptionalFloat: aFloat on: aStream
	"Print a symbolic representation of an exceptional Float value."

	aStream nextPutAll: (aFloat isNaN
				ifTrue: ['NaN']
				ifFalse: [aFloat > 0.0
						ifTrue: ['Infinity']
						ifFalse: ['-Infinity']])! !

!FloatPrinter methodsFor: 'printing' stamp: 'nice 6/10/2012 21:35'!
printFloat: aFloat on: aStream
	"Print aFloat on aStream in free format.
	The printed representation will have the minimum number of digits so that aFloat can be reconstructed exactly."
	| fixedFormat baseExpEstimate decPointCount |
	aFloat isFinite ifFalse: [^self printExceptionalFloat: aFloat on: aStream].
	self printSignOf: aFloat on: aStream.
	aFloat = 0.0 ifTrue: [aStream nextPutAll: '0.0'. ^ self].
	baseExpEstimate := self setFloat: aFloat.
	(fixedFormat := baseExpEstimate between: -3 and: 6)
		ifTrue:
			[(decPointCount := baseExpEstimate) <= 0
				ifTrue: [aStream nextPutAll: ('0.000' truncateTo: 2 - baseExpEstimate)]]
		ifFalse:
			[decPointCount := 1].
	decPointCount := self generateDigitsOn: aStream decPointCount: decPointCount.
	decPointCount > 0
		ifTrue:
			[aStream nextPutAll: ('00000.0' last: decPointCount +1)].
	fixedFormat
		ifFalse:
			[aStream nextPut: $e.
			baseExpEstimate - 1 printOn: aStream]! !

!FloatPrinter methodsFor: 'private-printing' stamp: ''!
printSignOf: aFloat on: aStream
	aFloat signBit = 1	ifTrue: [aStream nextPut: $-]! !

!FloatPrinter methodsFor: 'private' stamp: 'nice 6/8/2012 21:45'!
setFloat: aFloat
	"Prepare arithmetic interval for printing a Float.
	Answer the exponent in target print base"
	aFloat = 0.0
		ifTrue:
			[r := 0.
			s := mPlus := mMinus := 1.
			roundingLowIncludesLimits := roundingHighIncludesLimits := true.
			^1].
	^self estimateBaseExponent: (self setFloatInterval: aFloat)! !

!FloatPrinter methodsFor: 'private' stamp: 'nice 7/29/2012 23:05'!
setFloatInterval: f
	"Transform a Float into it's exact representation, decomposed in an integer significand and exp (biased to have an integer significand).
	Determine the interval from (r-mMinus)/s to (r+mPlus)/s where every number asFloat would result in this floating point absolute value r/s.
	The interval includes low bound if roundingLowIncludesLimits, and high bound if roundingHighIncludesLimits.
	Answer the unbiased exponent associated with normalized mantissa, because it will serve for estimating exponent in target print base.
	Note that bounds inclusion conditions are computed for IEEE default round to nearest even mode.
	Alternative rounding modes could be accomodated with slight variations of these conditions.
	Note that Float internal representation base is implicitely assumed to be 2 in this code.
	Note that care is taken to have interval variables represented with integer."
	
	| exp significand exponent |
	significand := f significandAsInteger.
	exponent := f exponent.
	exp := (exponent max: f class emin) + 1 - f class precision.
	roundingLowIncludesLimits :=  roundingHighIncludesLimits := significand even.
	exp >= 0
		ifTrue:
			[significand = 16r10000000000000
				ifTrue:
					[r := significand bitShift: 2 + exp.
					s := 4.
					mPlus := 2 * (mMinus := 1 bitShift: exp)]
				ifFalse:
					[r := significand bitShift: 1 + exp.
					s := 2.
					mPlus := mMinus := 1 bitShift: exp]]
		ifFalse:
			[(exponent > f class emin and: [significand = 16r10000000000000])
				ifTrue:
					[r := significand bitShift: 2.
					s := 1 bitShift: 2 - exp.
					mPlus := 2.
					mMinus := 1]
				ifFalse:
					[r := significand bitShift: 1.
					s := 1 bitShift: 1 - exp.
					mPlus := mMinus := 1]].
	^exponent! !

!FloatPrinter methodsFor: 'private' stamp: 'nice 6/10/2012 19:12'!
setFraction: aFraction
	"Prepare arithmetic interval for printing a Fraction.
	Answer the exponent in target print base"
	
	| baseExpEstimate scale |
	r := aFraction numerator.
	s := aFraction denominator.
	mPlus := mMinus := 0.
	roundingLowIncludesLimits := roundingHighIncludesLimits := true.
	r = 0 ifTrue: [^1].
	baseExpEstimate := r >= s
		ifTrue: [r // s numberOfDigitsInBase: base]
		ifFalse: [1 - (s // r numberOfDigitsInBase: base)].
	baseExpEstimate >= 0
		ifTrue: [s := s * (base raisedToInteger: baseExpEstimate)]
		ifFalse:
			[scale := base raisedToInteger: baseExpEstimate negated.
			r := r * scale.
			mPlus := mPlus * scale.
			mMinus := mMinus * scale].
	^((r + mPlus >= s) and: [roundingHighIncludesLimits or: [r + mPlus > s]])
		ifTrue: [baseExpEstimate + 1]
		ifFalse:
			[r := r * base.
			mPlus := mPlus * base.
			mMinus := mMinus * base.
			baseExpEstimate].! !

!FloatPrinter methodsFor: 'private' stamp: 'nice 6/1/2012 22:26'!
setTrailingDigitPower: anInteger
	"Adjust the bounds of the interval representing the float to be printed to account for rounding to prescribed power.
	This is used both when a fixed number of digits after fraction point or a fixed number of significant digits are specified."
	| delta |
	delta := s / 2 * (base raisedToInteger: anInteger).
	mMinus < delta
		ifTrue:
			[mMinus := delta.
			roundingLowIncludesLimits :=  true].
	mPlus < delta
		ifTrue:
			[mPlus := delta.
			roundingHighIncludesLimits := true]! !

!FloatPrinter methodsFor: 'private' stamp: 'nice 6/1/2012 22:22'!
simplifyFraction
	| slowbit den |
	(den := mPlus denominator lcm: mMinus denominator) = 1
		ifFalse: 
			[s := s * den.
			r := r * den.
			mPlus := mPlus * den.
			mMinus := mMinus * den].
	(slowbit := (s lowBit min: r lowBit) min: (mPlus lowBit min: mMinus lowBit)) > 1
		ifTrue:
			[slowbit := 1 - slowbit.
			s := s bitShift: slowbit.
			r := r bitShift: slowbit.
			mPlus := mPlus bitShift: slowbit.
			mMinus := mMinus bitShift: slowbit]! !

FloatPrinter subclass: #FloatPrinterForStore
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'NumberPrinter'!

!FloatPrinterForStore commentStamp: 'nice 5/23/2012 00:18' prior: 0!
A FloatPrinterForStore is specialized for storing a representation of a Float interpretable by Smalltalk Compiler.

It handles exceptional values with message send, and force printing of base radix when different from 10.
!

!FloatPrinterForStore methodsFor: 'private-printing' stamp: 'nice 6/2/2012 15:44'!
printExceptionalFloat: aFloat on: aStream
	"Print a symbolic representation of an exceptional Float value.
	Use a re-interpretable format."

	aStream nextPutAll: (aFloat isNaN
				ifTrue: ['Float nan']
				ifFalse: [aFloat > 0.0
						ifTrue: ['Float infinity']
						ifFalse: ['Float infinity negated']])! !

!FloatPrinterForStore methodsFor: 'private-printing' stamp: 'nice 5/31/2012 22:40'!
printSignOf: aFloat on: aStream
	super printSignOf: aFloat on: aStream.
	base = 10
		ifFalse:
			[base printOn: aStream.
			aStream nextPut: $r]! !

FloatPrinter subclass: #VersatileFloatPrinter
	instanceVariableNames: 'fractionSeparator digitCount exponentChar totalWidth padWithZeros exponentDigitCount showTrailingZeros alwaysPrintFractionPart alwaysPrintFractionSeparator positiveSignChar forceExponentSign'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'NumberPrinter'!

!VersatileFloatPrinter commentStamp: 'nice 9/28/2021 01:03' prior: 0!
A VersatileFloatPrinter is class specialized in printing floating point numbers with versatile format.
It provides many parameters to do so.
Subclasses implement various formats using these parameters.

Instance Variables
	alwaysShowFractionSeparator:	<Boolean>
	alwaysShowFractionPart:	<Boolean>
	digitCount		<Integer | nil>
	exponentChar:		<Character>
	exponentDigitCount:		<Integer>
	forceExponentSign	<Boolean>
	fractionSeparator:		<Character>
	padWithZeros:	<Boolean>
	positiveSignChar:		<Character | nil>
	showtrailingZeros:		<Boolean>
	totalWidth:		<Integer | nil>

alwaysShowFractionSeparator
	- a fraction separator is always printed when true

alwaysShowFractionPart
	- a fraction separator and at least one digit of fraction part are always printed when true

digitCount
	- if not nil, specify a restricted number of significant digits to be used

exponentChar
	- the character marking the separation of exponent part

exponentDigitCount
	- the minimum number of digits to be printed in exponent

padWithZeros
	- when a width is specified, the output is padded with zeros if padWithZeros is true, with spaces if false 

positiveSignChar
	- the character to be printed before positive numbers, nil if none

forceExponentSign
	- when true, print a + character before positive exponents
		
fractionSeparator
	- the character marking the separation between integer and fraction parts

showtrailingZeros
	- if true, always print the trailing zeros when a digitCount was specified and greater than strictly necessary digits
	  if false, don't print the trailing zeros, and don't even print the fractionSeparator if all fractional digits are zero

totalWidth
	- if not nil, the specifies the total width of printed output.
	If printed representation is smaller than this width, it is left padded with zeros or spaces (depending on padWithZeros instance variable).
	If it is larger, then the printed representation is not restricted to prescribed width.!

VersatileFloatPrinter subclass: #FloatPrinterFixedFormat
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'NumberPrinter'!

!FloatPrinterFixedFormat commentStamp: 'nice 5/29/2012 22:25' prior: 0!
A FloatPrinterFixedFormat is specialized for printing a Float in fixed format.

In fixed format, the digitCount instance variable is specifying the number of places to be used AFTER the fractionSeparator.
In fixed format, no exponent scaling is used.
!

!FloatPrinterFixedFormat methodsFor: 'initialize-release' stamp: 'nice 9/28/2021 01:14'!
actAsPrintf
	"setup to mimic C printf default behavior"
	
	super actAsPrintf.
	showTrailingZeros := true! !

!FloatPrinterFixedFormat methodsFor: 'private-printing' stamp: 'nice 6/10/2012 18:28'!
baseExpLimit: baseExpEstimate
	"In fixed format, digitCount specifies the number of digits AFTER fractionSeparator."
	
	showTrailingZeros ifTrue: [digitCount ifNotNil: [digitCount > 0 ifTrue: [^1 - digitCount]]].
	^1! !

!FloatPrinterFixedFormat methodsFor: 'printing' stamp: 'nice 7/29/2012 23:06'!
print: aNumber limitedWidthOn: aStream
	"Print aNumber on aStream with fixed format and limited width."
	| size decPointCount baseExpEstimate |
	aNumber isFinite ifFalse: [^self printExceptionalFloat: aNumber limitedWidthOn: aStream].
	size := self fixedFormatSizeOf: aNumber.
	baseExpEstimate := aNumber initializePrinter: self.
	padWithZeros ifFalse: [self printSpaces: totalWidth - size on: aStream].
	self printSignOf: aNumber on: aStream.
	padWithZeros ifTrue: [self printZeros: totalWidth - size on: aStream].
	(decPointCount := baseExpEstimate) <= 0
		ifTrue:
			[aStream nextPut: $0.
			digitCount ifNotNil: [digitCount + baseExpEstimate <= 0
				ifTrue:
					[decPointCount := 1.
					^self printTrailingFractionPartZerosOn: aStream decPointCount: decPointCount baseExpEstimate: baseExpEstimate]].
			self printLeadingFractionPartZerosOn: aStream baseExpEstimate: baseExpEstimate].
	decPointCount := self generateDigitsOn: aStream decPointCount: decPointCount.
	decPointCount := self printTrailingIntegerPartZerosOn: aStream decPointCount: decPointCount.
	^self printTrailingFractionPartZerosOn: aStream decPointCount: decPointCount baseExpEstimate: baseExpEstimate! !

!FloatPrinterFixedFormat methodsFor: 'printing' stamp: 'nice 7/29/2012 22:39'!
print: aNumber on: aStream
	"Print a float on a stream in fixed format.
	The number of digits after floating point is specified by digitCount.
	Beyond Float precision, zeroes are appended.
	When showTrailingZeros is false, the trailing zeroes after floating point will be omitted.
	If all fractional digits are zeros, the floating point is omitted too."
	| decPointCount baseExpEstimate |
	totalWidth ifNotNil: [^self print: aNumber limitedWidthOn: aStream].
	aNumber isFinite ifFalse: [^self printExceptionalFloat: aNumber on: aStream].
	self printSignOf: aNumber on: aStream.
	baseExpEstimate := aNumber initializePrinter: self.
	(decPointCount := baseExpEstimate) <= 0
		ifTrue:
			[aStream nextPut: $0.
			digitCount ifNotNil: [digitCount + baseExpEstimate <= 0
				ifTrue:
					[decPointCount := 1.
					^self printTrailingFractionPartZerosOn: aStream decPointCount: decPointCount baseExpEstimate: baseExpEstimate]].
			self printLeadingFractionPartZerosOn: aStream baseExpEstimate: baseExpEstimate].
	decPointCount := self generateDigitsOn: aStream decPointCount: decPointCount.
	decPointCount := self printTrailingIntegerPartZerosOn: aStream decPointCount: decPointCount.
	^self printTrailingFractionPartZerosOn: aStream decPointCount: decPointCount baseExpEstimate: baseExpEstimate! !

!FloatPrinterFixedFormat methodsFor: 'private' stamp: 'nice 6/2/2012 16:11'!
setFloat: aFloat
	| base2Exponent |
	aFloat = 0.0 ifTrue: [^super setFloat: aFloat].
	base2Exponent := self setFloatInterval: aFloat.
	digitCount ifNotNil: [self setTrailingDigitPower: 0 - digitCount].
	^self estimateBaseExponent: base2Exponent! !

!FloatPrinterFixedFormat methodsFor: 'private' stamp: 'nice 6/10/2012 21:25'!
setFraction: aFraction
	| baseExpEstimate scale |
	r := aFraction numerator.
	s := aFraction denominator.
	mPlus := mMinus := 0.
	roundingLowIncludesLimits := roundingHighIncludesLimits := true.
	r = 0 ifTrue: [^1].
	baseExpEstimate := r >= s
		ifTrue: [r // s numberOfDigitsInBase: base]
		ifFalse: [1 - (s // r numberOfDigitsInBase: base)].
	self setTrailingDigitPower: 0 - (digitCount ifNil: [6]).
	baseExpEstimate >= 0
		ifTrue: [s := s * (base raisedToInteger: baseExpEstimate)]
		ifFalse:
			[scale := base raisedToInteger: baseExpEstimate negated.
			r := r * scale.
			mPlus := mPlus * scale.
			mMinus := mMinus * scale].
	^((r + mPlus >= s) and: [roundingHighIncludesLimits or: [r + mPlus > s]])
		ifTrue: [baseExpEstimate + 1]
		ifFalse:
			[r := r * base.
			mPlus := mPlus * base.
			mMinus := mMinus * base.
			baseExpEstimate].! !

VersatileFloatPrinter subclass: #FloatPrinterFreeFormat
	instanceVariableNames: 'fixedFormat'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'NumberPrinter'!

!FloatPrinterFreeFormat commentStamp: 'nice 3/5/2018 23:52' prior: 0!
A FloatPrinterFreeFormat is specialized for printing a Float in free format.

Free format uses either scientific format or fixed format, depending on the number scale.
The rules are
- no more than 6 digits left of fraction separator if magnitude is greater than or equal to one,
- and no more than 3 leading fractional zeros if magnitude is smaller than one.

If a digitCount is specified, it represent the TOTAL number of significant digits, not only the fractional digits.
In this case, the rules for choosing between fixed and scientific format are different.
- no more than digitCount digits left of fraction separator if magnitude is greater than or equal to one,
- scientific is systematically used for magnitude less than one (except for zero)

This last property differs from C printf format '%g' which keep the 3 leading fractional zeros rule when a digitCount is specified.

If totalWidth is specified, then rules are again different: the shortest representation is chosen.

Instance Variables
	fixedFormat:	<Boolean>
		
fixedFormat
	- indicate whether fixed format is used if true, or scientific format if false
!

!FloatPrinterFreeFormat methodsFor: 'private-printing' stamp: 'nice 6/10/2012 18:28'!
baseExpLimit: baseExpEstimate
	"In case of free format, the digits printed left of fractionSeparator count"

	showTrailingZeros ifTrue: [digitCount ifNotNil: [^fixedFormat
			ifTrue: [(baseExpEstimate max: 1) + 1 - digitCount]
			ifFalse: [2 - digitCount]]].
	^1! !

!FloatPrinterFreeFormat methodsFor: 'printing' stamp: 'nice 7/29/2012 22:38'!
print: aNumber limitedWidthOn: aStream
	"Print aNumber on aStream with free format and limited width."
	
	| fs size ss decPointCount baseExpEstimate |
	aNumber isFinite ifFalse: [^self printExceptionalFloat: aNumber limitedWidthOn: aStream].
	fixedFormat := true.
	fs := self fixedFormatSizeOf: aNumber.
	fixedFormat := false.
	ss := self scientificFormatSizeOf: aNumber.
	baseExpEstimate := aNumber initializePrinter: self.
	fixedFormat := fs <= ss.
	size := fixedFormat ifTrue: [fs] ifFalse: [ss].
	padWithZeros ifFalse: [self printSpaces: totalWidth - size on: aStream].
	self printSignOf: aNumber on: aStream.
	padWithZeros ifTrue: [self printZeros: totalWidth - size on: aStream].
	(decPointCount := fixedFormat
		ifTrue: [baseExpEstimate]
		ifFalse: [1]) <= 0
			ifTrue:
				[aStream nextPut: $0.
				self printLeadingFractionPartZerosOn: aStream baseExpEstimate: baseExpEstimate].
	decPointCount := self generateDigitsOn: aStream decPointCount: decPointCount.
	decPointCount := self printTrailingIntegerPartZerosOn: aStream decPointCount: decPointCount.
	self printTrailingFractionPartZerosOn: aStream decPointCount: decPointCount baseExpEstimate: baseExpEstimate.
	fixedFormat ifFalse: [self printExponent: baseExpEstimate - 1 on: aStream]! !

!FloatPrinterFreeFormat methodsFor: 'printing' stamp: 'nice 7/29/2012 22:38'!
print: aNumber on: aStream
	"Print aNumber on aStream in free format.
	The printed representation will have the minimum number of digits so that aNumber can be reconstructed exactly."
	| decPointCount baseExpEstimate |
	totalWidth ifNotNil: [^self print: aNumber limitedWidthOn: aStream].
	aNumber isFinite ifFalse: [^self printExceptionalFloat: aNumber on: aStream].
	self printSignOf: aNumber on: aStream.
	baseExpEstimate := aNumber initializePrinter: self.
	fixedFormat := digitCount
		ifNil: [baseExpEstimate between: -3 and: 6]
		ifNotNil: [baseExpEstimate between: 1 and: digitCount].
	(decPointCount := fixedFormat
		ifTrue: [baseExpEstimate]
		ifFalse: [1]) <= 0
			ifTrue:
				[aStream nextPut: $0.
				self printLeadingFractionPartZerosOn: aStream baseExpEstimate: baseExpEstimate].
	decPointCount := self generateDigitsOn: aStream decPointCount: decPointCount.
	decPointCount := self printTrailingIntegerPartZerosOn: aStream decPointCount: decPointCount.
	self printTrailingFractionPartZerosOn: aStream decPointCount: decPointCount baseExpEstimate: baseExpEstimate.
	fixedFormat ifFalse: [self printExponent: baseExpEstimate - 1 on: aStream]! !

!FloatPrinterFreeFormat methodsFor: 'private' stamp: 'nice 6/8/2012 21:45'!
setFloat: aFloat
	| baseExpEstimate |
	aFloat = 0.0 ifTrue: [^super setFloat: aFloat].
	baseExpEstimate := self estimateBaseExponent: (self setFloatInterval: aFloat).
	digitCount
		ifNotNil:
			[baseExpEstimate := self adjustEstimatedBaseExponent: baseExpEstimate forRoundingToDigitCount: digitCount.
			self setTrailingDigitPower: 1 - digitCount].
	^baseExpEstimate! !

!FloatPrinterFreeFormat methodsFor: 'private' stamp: 'nice 6/10/2012 21:28'!
setFraction: aFraction
	| baseExpEstimate d |
	baseExpEstimate := super setFraction: aFraction.
	d := digitCount ifNil: [6].
	baseExpEstimate := self adjustEstimatedBaseExponent: baseExpEstimate forRoundingToDigitCount: d.
	self setTrailingDigitPower: 1 - d.
	^baseExpEstimate! !

!FloatPrinterFreeFormat methodsFor: 'private-measuring' stamp: ''!
trueFixedFractionCount: baseExpEstimate
	"Answer how many digits are necessary, after the fraction separator"

	| fractionCount |
	fractionCount := digitCount isNil 
		ifTrue: [self copy countSignificandDigits - baseExpEstimate]
		ifFalse: [showTrailingZeros
			ifTrue: [1 - (self baseExpLimit: baseExpEstimate)]
			ifFalse: [self copy countSignificandDigits - baseExpEstimate]].
	fractionCount := fractionCount max: (alwaysPrintFractionPart
		ifTrue: [1]
		ifFalse: [0]).
	^fractionCount! !

!FloatPrinterFreeFormat methodsFor: 'private-measuring' stamp: ''!
trueScientificFractionCount
	"Answer how many digits are necessary, after the fraction separator"

	| fractionCount |
	fractionCount := (digitCount notNil and: [showTrailingZeros])
		ifTrue: [digitCount - 1]
		ifFalse: [self copy countSignificandDigits - 1].
	alwaysPrintFractionPart ifTrue: [fractionCount := fractionCount max: 1].
	^fractionCount max: 0! !

VersatileFloatPrinter subclass: #FloatPrinterHexadecimalFormat
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'NumberPrinter'!

!FloatPrinterHexadecimalFormat commentStamp: 'nice 9/27/2021 22:05' prior: 0!
A FloatPrinterHexadecimalFormat will print a Float in C compatible printf %a format

In this format, the output if formed of following sequence:
- optional spaces if width is specified with space padding
- a sign (optional if +)
- followed by padding zeroes if width is specified with 0 padding
- followed by is a prefix '0x'
- followed by the implied bit - which is either 1 (for normal) or 0 (for denormal), or 2 and 1 if the rounding overflows
- followed by fraction separator (character $.) - optional if fraction part is zero
- followed by fraction part of significand (52 bits for IEEE 754 double precision make at most 13 hexadecimal digits)
- followed by binary exponent separator (character $p)
- followed by binary exponent

Instance Variables
!

!FloatPrinterHexadecimalFormat methodsFor: 'initialize-release' stamp: 'nice 9/28/2021 01:14'!
actAsPrintf
	"setup to mimic C printf default behavior"
	
	super actAsPrintf.
	showTrailingZeros := true! !

!FloatPrinterHexadecimalFormat methodsFor: 'initialize-release' stamp: 'nice 9/27/2021 22:06'!
initialize
	super initialize.
	exponentChar := $p.! !

!FloatPrinterHexadecimalFormat methodsFor: 'private-printing' stamp: 'nice 9/27/2021 22:44'!
numberOfFractionDigitsHi: hi lo: lo
	"Answer with the number of fraction digits to be printed,
	given the hi (20 bits) and lo (32 bits) part of fraction bits."
	| fractionCount |
	fractionCount := showTrailingZeros
				ifTrue: [digitCount ifNil: [13]]
				ifFalse: [lo > 0
						ifTrue: [13 - (lo lowBit - 1 // 4)]
						ifFalse: [hi > 0
								ifTrue: [5 - (hi lowBit - 1 // 4)]
								ifFalse: [0]]].
	alwaysPrintFractionPart
		ifTrue: [fractionCount := fractionCount max: 1].
	^fractionCount! !

!FloatPrinterHexadecimalFormat methodsFor: 'printing' stamp: 'nice 9/27/2021 22:19'!
print: aFloat limitedWidthOn: aStream
	"Print aNumber on aStream with hexa-decimal format and limited width."
	
	| hi lo fractionCount exp hilo integerPart requiredWidth |
	aFloat isFinite
		ifFalse: [^ self printExceptionalFloat: aFloat on: aStream].
	hilo := self roundedFractionBitsForFloat: aFloat.
	hi := hilo first.	"highest 20 bits or fraction part (or 21 bits if rounding did denormalize)"
	lo := hilo last	"lowest 32 bits of fraction part".
	integerPart := ((aFloat isDenormal or: [aFloat isZero])
				ifTrue: [0]
				ifFalse: [1])
				+ (hi >> 20 "Handle un-normalizing rounding").
	fractionCount := self numberOfFractionDigitsHi: (hi bitAnd: 16rFFFFF) lo: lo.
	exp := aFloat isZero
				ifTrue: [0]
				ifFalse: [aFloat isDenormal
						ifTrue: [Float emin]
						ifFalse: [aFloat exponent]].
	requiredWidth := ((self forceSign or: [aFloat signBit = 1]) ifTrue: [1] ifFalse: [0])
		+ 2 "0x" + 1 "integerPart" + fractionCount + 2 "exponent letter + sign" + (exp numberOfDigitsInBase: 10).
	(alwaysPrintFractionSeparator
			or: [fractionCount > 0])
		ifTrue: [requiredWidth := requiredWidth + 1].
	(padWithZeros or: [totalWidth <= requiredWidth])
		ifFalse: [self printSpaces: totalWidth - requiredWidth on: aStream].
	self printSignOf: aFloat on: aStream.
	aStream nextPutAll: '0x'.
	(padWithZeros and: [totalWidth > requiredWidth])
		ifTrue: [self printZeros: totalWidth - requiredWidth on: aStream].
	integerPart printOn: aStream.
	(alwaysPrintFractionSeparator
			or: [fractionCount > 0])
		ifTrue: [aStream nextPut: fractionSeparator].
	self printFractionOn: aStream digitCount: fractionCount hi: (hi bitAnd: 16rFFFFF) lo: lo.
	self printExponentOfFloat: aFloat on: aStream.! !

!FloatPrinterHexadecimalFormat methodsFor: 'printing' stamp: 'nice 9/27/2021 21:59'!
print: aFloat on: aStream 
	"Print a float on a stream in hexadecimal format.
	The number of digits after floating point is specified by digitCount (if notNil).
	Beyond Float precision, zeroes are appended (if requested).
	When showTrailingZeros is false, the trailing zeroes after floating point will be omitted.
	If all fractional digits are zeros, the fraction separator is omitted too unless requested."
	| hi lo fractionCount hilo integerPart |
	totalWidth
		ifNotNil: [^ self print: aFloat limitedWidthOn: aStream].
	aFloat isFinite
		ifFalse: [^ self printExceptionalFloat: aFloat on: aStream].
	hilo := self roundedFractionBitsForFloat: aFloat.
	hi := hilo first.	"highest 20 bits or fraction part (or 21 bits if rounding did denormalize)"
	lo := hilo last	"lowest 32 bits of fraction part".
	integerPart := ((aFloat isDenormal or: [aFloat isZero])
				ifTrue: [0]
				ifFalse: [1])
				+ (hi >> 20 "Handle un-normalizing rounding").
	"Make sure to renormalize the fraction bits, and count fractional digits"
	hi := hi bitAnd: 16rFFFFF.
	fractionCount := self numberOfFractionDigitsHi: hi lo: lo.
	"Now proceed with printing"
	self printSignOf: aFloat on: aStream.
	aStream nextPutAll: '0x'.
	integerPart printOn: aStream.
	(alwaysPrintFractionSeparator
			or: [fractionCount > 0])
		ifTrue: [aStream nextPut: fractionSeparator].
	self printFractionOn: aStream digitCount: fractionCount hi: hi lo: lo.
	self printExponentOfFloat: aFloat on: aStream.! !

!FloatPrinterHexadecimalFormat methodsFor: 'private-printing' stamp: 'nice 9/27/2021 21:55'!
printExponentOfFloat: aFloat on: aStream
	"print the binary exponent"
	
	| exp |
	exp := aFloat isZero
				ifTrue: [0]
				ifFalse: [aFloat isDenormal
						ifTrue: [Float emin]
						ifFalse: [aFloat exponent]].
	self printExponent: exp on: aStream! !

!FloatPrinterHexadecimalFormat methodsFor: 'private-printing' stamp: 'nice 9/28/2021 01:12'!
printFractionOn: aStream digitCount: fractionCount hi: hi lo: lo
	"Print prescribed number of digits composing the fraction part of this float
	given the hi (20 bits) and lo (32 bits) part of fraction bits."

	fractionCount > 0 ifFalse: [^self].
	self printZeros: (fractionCount min: 5 - (hi highBit + 3 // 4)) on: aStream.
	hi > 0 ifTrue: [hi >> (5 - (fractionCount min: 5) * 4) printOn: aStream base: 16].
	fractionCount > 5
		ifTrue:
			[self printZeros: (fractionCount - 5 min: 8 - (lo highBit + 3 // 4)) on: aStream.
			lo > 0 ifTrue: [lo >> (8 - (fractionCount - 5 min: 8) * 4) printOn: aStream base: 16]].
	fractionCount > 13
		ifTrue:
			[self printZeros: fractionCount - 13 on: aStream].! !

!FloatPrinterHexadecimalFormat methodsFor: 'private-printing' stamp: 'nice 9/27/2021 21:54'!
roundedFractionBitsForFloat: aFloat
	"Extract the fraction bits of aFloat from its significand.
	The implied 1 bit in IEEE 754 float representation will be printed as the integer part and is thus ignored.
	Handle the remaining 52bits significand corresponding to fraction bits.
	We know that it fits on 13 hexadecimal digits.
	Split those bits in two parts of 5 hex highest digits (20 bits) and 8 hex lowest digits (32 bits).
	Round this significand to nearest, tie to even if digitCount is limited.
	Beware, rounding can un-normalize the significand to 53 bits (that is hi highBit = 21).
	In such case, the integerPart will be incremented."
	| hi lo |
	hi := (aFloat basicAt: 1) bitAnd: 16rFFFFF.
	lo := aFloat basicAt: 2.
	digitCount
		ifNotNil: [digitCount < 13
				ifTrue:
					[| fraction shift rem rounded |
					fraction := hi << 32 + lo.
					shift := 13 - digitCount * 4.
					rounded := fraction >> shift << shift.
					rem := fraction - rounded.
					rem highBit = shift
						ifTrue: [rem lowBit < shift
								ifTrue: [rounded := rounded + (1 << shift)]
								ifFalse: ["Tie to even"
									rounded := rounded + (rounded bitAnd: 1 << shift)]].
					hi := rounded >> 32.
					lo := rounded bitAnd: 16rFFFFFFFF]].
	^Array with: hi with: lo! !

VersatileFloatPrinter subclass: #FloatPrinterScientificFormat
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'NumberPrinter'!

!FloatPrinterScientificFormat commentStamp: 'nice 4/8/2016 00:12' prior: 0!
A FloatPrinterScientificFormat is specialized for printing a Float in scientific format.

In scientific format,
- a single digit is printed before the fractionSeparator,
- an exponent scaling is always printed.

The number of fractional digits printed after the fractionSeparator is controlled this way:
- if digitCount is specified and > 0,
	* if showTrailingZeros is true then exactly digitCount digits are printed after the fractionSeparator
	* else if alwaysPrintFractionPart is true, at least one digit is printed
	* else if alwaysPrintFractionSeparator is true, at least the fractionSeparator is printed, followed by strictly necessary digits
	* else the fractionSeparator is printed only if there is more than zero strictly necessary digits.
- if digitCount is specified and = 0, no fractionSeparator and no fractional digits are printed, unless alwaysPrintFractionPart or alwaysPrintFractionSeparator is true
- if digitCount is unspecified (nil), then the required number of fractional digits are printed so that the float can be reconstructed exactly. At least 1 fractional digit is always printed if alwaysPrintFractionPart is true.
!

!FloatPrinterScientificFormat methodsFor: 'initialize-release' stamp: 'nice 9/28/2021 01:14'!
actAsPrintf
	"setup to mimic C printf default behavior"
	
	super actAsPrintf.
	showTrailingZeros := true! !

!FloatPrinterScientificFormat methodsFor: 'private-printing' stamp: 'nice 6/10/2012 18:28'!
baseExpLimit: baseExpEstimate
	"In scientific format, digitCount specifies the number of digits AFTER fractionSeparator."
	
	showTrailingZeros ifTrue: [digitCount ifNotNil: [digitCount > 0 ifTrue: [^1 - digitCount]]].
	^1! !

!FloatPrinterScientificFormat methodsFor: 'printing' stamp: 'nice 6/10/2012 21:39'!
print: aNumber limitedWidthOn: aStream
	"Print aNumber on aStream with scientific format and limited width."
	| size decPointCount baseExpEstimate |
	aNumber isFinite ifFalse: [^self printExceptionalFloat: aNumber limitedWidthOn: aStream].
	size := self scientificFormatSizeOf: aNumber.
	decPointCount := 1.
	padWithZeros ifFalse: [self printSpaces: totalWidth - size on: aStream].
	self printSignOf: aNumber on: aStream.
	padWithZeros ifTrue: [self printZeros: totalWidth - size on: aStream].
	baseExpEstimate := aNumber initializePrinter: self.
	decPointCount := self generateDigitsOn: aStream decPointCount: decPointCount.
	self printTrailingFractionPartZerosOn: aStream decPointCount: decPointCount baseExpEstimate: baseExpEstimate.
	self printExponent: baseExpEstimate - 1 on: aStream! !

!FloatPrinterScientificFormat methodsFor: 'printing' stamp: 'nice 6/10/2012 21:39'!
print: aNumber on: aStream
	"Print aNumber on aStream in scientific notation."
	| decPointCount baseExpEstimate |
	totalWidth ifNotNil: [^self print: aNumber limitedWidthOn: aStream].
	aNumber isFinite ifFalse: [^self printExceptionalFloat: aNumber on: aStream].
	decPointCount := 1.
	self printSignOf: aNumber on: aStream.
	baseExpEstimate := aNumber initializePrinter: self.
	decPointCount := self generateDigitsOn: aStream decPointCount: decPointCount.
	self printTrailingFractionPartZerosOn: aStream decPointCount: decPointCount baseExpEstimate: baseExpEstimate.
	self printExponent: baseExpEstimate - 1 on: aStream! !

!FloatPrinterScientificFormat methodsFor: 'private' stamp: 'nice 6/8/2012 21:45'!
setFloat: aFloat
	| baseExpEstimate |
	aFloat = 0.0 ifTrue: [^super setFloat: aFloat].
	baseExpEstimate := self estimateBaseExponent: (self setFloatInterval: aFloat).
	digitCount
		ifNotNil:
			[baseExpEstimate := self adjustEstimatedBaseExponent: baseExpEstimate forRoundingToDigitCount: 1 + digitCount.
			self setTrailingDigitPower: 0 - digitCount].
	^baseExpEstimate! !

!FloatPrinterScientificFormat methodsFor: 'private' stamp: 'nice 6/10/2012 21:27'!
setFraction: aFraction
	| baseExpEstimate d |
	baseExpEstimate := super setFraction: aFraction.
	d := digitCount ifNil: [6].
	baseExpEstimate := self adjustEstimatedBaseExponent: baseExpEstimate forRoundingToDigitCount: 1 + d.
	self setTrailingDigitPower: 0 - d.
	^baseExpEstimate! !

VersatileFloatPrinter subclass: #FloatPrinterThousandFormat
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'NumberPrinter'!

!FloatPrinterThousandFormat commentStamp: 'nice 4/7/2016 20:17' prior: 0!
A FloatPrinterThousandFormat is specialized for printing a Float with exponents multiple of 3.

In thousand format,
- 1 to 3 digits is printed before the fractionSeparator,
- an exponent scaling is always printed if different from zero, but allways multiple of 3.

Examples:
100.2  10.34e3 1.05e6  -33.4e-6

- if digitCount is > 0,  it controls the total number of significative figures (BEFORE+AFTER fraction separator)
  In this case, the number of fractional digits printed after the fractionSeparator will depend both on exponent and digitCounts
- if digitCount <= 0, then it directly controls the number of digits AFTER the fraction separator.
- if digitCount is unspecified (nil), then the required number of fractional digits are printed so that the float can be reconstructed exactly.

If showTrailingZeros is true then all fractional digits are considered strictly necessary.
If alwaysPrintFractionPart is true, then at least 1 fractional digit is considered strictly necessary.
Else the trailing zeroes are skipped

The fractionSeparator is printed only if there is more than zero strictly necessary digits or if alwaysPrintFractionSeparator is true.
!

!FloatPrinterThousandFormat methodsFor: 'private-printing' stamp: 'nice 4/7/2016 21:16'!
baseExpLimit: baseExpEstimate
	"In thousand format, digitCount specifies the number of digits AFTER fractionSeparator if <= 0."
	
	showTrailingZeros ifTrue: [digitCount ifNotNil: [digitCount > 0
			ifTrue: [^(baseExpEstimate - 1 \\ 3) + 2 - digitCount]
			ifFalse: [^1 + digitCount]]].
	^1! !

!FloatPrinterThousandFormat methodsFor: 'printing' stamp: 'nice 4/7/2016 22:25'!
print: aNumber limitedWidthOn: aStream
	"Print aNumber on aStream with scientific format and limited width."
	| size decPointCount baseExpEstimate exp |
	aNumber isFinite ifFalse: [^self printExceptionalFloat: aNumber limitedWidthOn: aStream].
	size := self thousandFormatSizeOf: aNumber.
	baseExpEstimate := aNumber initializePrinter: self.
	decPointCount := baseExpEstimate - 1 \\ 3 + 1.
	padWithZeros
		ifTrue: [self printSpaces: (totalWidth - size - 3 + decPointCount) on: aStream]
		ifFalse: [self printSpaces: totalWidth - size on: aStream].
	self printSignOf: aNumber on: aStream.
	padWithZeros ifTrue: [self printZeros: (3 - decPointCount min: totalWidth - size) on: aStream].
	decPointCount := self generateDigitsOn: aStream decPointCount: decPointCount.
	decPointCount := self printTrailingIntegerPartZerosOn: aStream decPointCount: decPointCount.
	self printTrailingFractionPartZerosOn: aStream decPointCount: decPointCount baseExpEstimate: baseExpEstimate.
	exp := baseExpEstimate - 1 // 3 * 3.
	exp isZero ifFalse: [self printExponent: exp on: aStream]! !

!FloatPrinterThousandFormat methodsFor: 'printing' stamp: 'nice 4/7/2016 21:47'!
print: aNumber on: aStream
	"Print aNumber on aStream in scientific notation."
	| decPointCount baseExpEstimate exp |
	totalWidth ifNotNil: [^self print: aNumber limitedWidthOn: aStream].
	aNumber isFinite ifFalse: [^self printExceptionalFloat: aNumber on: aStream].
	self printSignOf: aNumber on: aStream.
	baseExpEstimate := aNumber initializePrinter: self.
	decPointCount := baseExpEstimate - 1 \\ 3 + 1.
	decPointCount := self generateDigitsOn: aStream decPointCount: decPointCount.
	decPointCount := self printTrailingIntegerPartZerosOn: aStream decPointCount: decPointCount.
	self printTrailingFractionPartZerosOn: aStream decPointCount: decPointCount baseExpEstimate: baseExpEstimate.
	exp := baseExpEstimate - 1 // 3 * 3.
	exp isZero ifFalse: [self printExponent: exp on: aStream]! !

!FloatPrinterThousandFormat methodsFor: 'private' stamp: 'nice 4/8/2016 00:08'!
setFloat: aFloat
	| baseExpEstimate |
	aFloat = 0.0 ifTrue: [^super setFloat: aFloat].
	baseExpEstimate := self estimateBaseExponent: (self setFloatInterval: aFloat).
	digitCount
		ifNotNil:
			[baseExpEstimate := self adjustEstimatedBaseExponent: baseExpEstimate forRoundingToDigitCount: (digitCount > 0
				ifTrue: [digitCount] ifFalse: [baseExpEstimate - 1 \\ 3 + 1 - digitCount]).
			self setTrailingDigitPower: (digitCount > 0 ifTrue: [1 - digitCount] ifFalse: [digitCount - (baseExpEstimate - 1 \\ 3)])].
	^baseExpEstimate! !

!FloatPrinterThousandFormat methodsFor: 'private' stamp: 'nice 4/8/2016 00:13'!
setFraction: aFraction
	| baseExpEstimate d |
	baseExpEstimate := super setFraction: aFraction.
	d := digitCount ifNil: [6].
	baseExpEstimate := self adjustEstimatedBaseExponent: baseExpEstimate forRoundingToDigitCount: (d > 0
				ifTrue: [d] ifFalse: [baseExpEstimate - 1 \\ 3 + 1 - d]).
	self setTrailingDigitPower: (d > 0 ifTrue: [1 - d] ifFalse: [d - (baseExpEstimate - 1 \\ 3)]).
	^baseExpEstimate! !

!VersatileFloatPrinter methodsFor: 'initialize-release' stamp: 'nice 9/28/2021 01:15'!
actAsPrintf
	"setup to mimic C printf default behavior"
	
	digitCount := 6.
	exponentDigitCount := 2.
	forceExponentSign := true.! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 5/29/2012 23:43'!
alwaysPrintFractionPart: aBoolean
	alwaysPrintFractionPart := aBoolean! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 5/29/2012 23:43'!
alwaysPrintFractionSeparator: aBoolean
	alwaysPrintFractionSeparator := aBoolean! !

!VersatileFloatPrinter methodsFor: 'private-printing' stamp: ''!
baseExpLimit: baseExpEstimate
	"Encode how many digits at least are expected after fraction separator.
	The same convention as baseExpEstimate is used.
	That is 0 for 1 digit after fractionSeparator, -1 for two digits...
	A positive value means no fractional digits are required."
	
	^self subclassResponsibility! !

!VersatileFloatPrinter methodsFor: 'private-measuring' stamp: 'nice 6/10/2012 18:44'!
countSignificandDigits
	"Count how many digits are necessary to print at required precision.
	This is not including eventual leading nor trailing zeros."
	| d shead slowbit size |
	self simplifyFraction.
	slowbit := 1 - s lowBit .
	shead := s bitShift: slowbit.
	size := 1.
	[d := (r bitShift: slowbit) // shead.
	r := r - (d * s).
	((r <= mMinus) and: [roundingLowIncludesLimits or: [r < mMinus]]) or:
	[(r + mPlus >= s) and: [roundingHighIncludesLimits or: [r + mPlus > s]]]] whileFalse:
		[size := size + 1.
		r := r * base.
		mPlus := mPlus * base.
		mMinus := mMinus * base].
	^size! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 5/22/2012 23:44'!
digitCount: anIntegerOrNil
	digitCount := anIntegerOrNil! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 5/24/2012 22:47'!
exponentChar: aCharacter
	"Set the character used to separate exponent part."
	exponentChar := aCharacter! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 5/28/2012 21:48'!
exponentDigitCount: anInteger
	"Set the minimum number of digits used in exponent"
	exponentDigitCount := anInteger! !

!VersatileFloatPrinter methodsFor: 'private-measuring' stamp: ''!
fixedFormatSizeOf: aNumber
	"Answer the size required for printing aNumber in fixed format."

	| size integerCount fractionCount baseExpEstimate |
	baseExpEstimate := aNumber initializePrinter: self.

	"count sign"
	size := (self forceSign or: [aNumber signBit = 1]) ifTrue: [1] ifFalse: [0].

	"count the digits"
	integerCount := self trueFixedIntegerCount: baseExpEstimate.
	fractionCount := self trueFixedFractionCount: baseExpEstimate.

	"count the fractionSeparator"
	(fractionCount > 0 or: [alwaysPrintFractionSeparator]) ifTrue: [size := size + 1].
	
	^size + integerCount + fractionCount! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 9/28/2021 01:00'!
forceExponentSign: aBoolean
	forceExponentSign := aBoolean! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 6/2/2012 20:33'!
forceSign
	^positiveSignChar notNil! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: ''!
forceSign: aBoolean
	positiveSignChar := $+! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 6/1/2012 22:25'!
fractionSeparator
	^fractionSeparator! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 5/22/2012 21:44'!
fractionSeparator: aCharacter
	"Set the character used to separate integer part from fraction part."
	fractionSeparator := aCharacter! !

!VersatileFloatPrinter methodsFor: 'initialize-release' stamp: 'nice 9/28/2021 01:01'!
initialize
	super initialize.
	fractionSeparator := $..
	exponentChar := $e.
	exponentDigitCount := 1.
	padWithZeros := false.
	showTrailingZeros := false.
	alwaysPrintFractionPart := false.
	alwaysPrintFractionSeparator := false.
	forceExponentSign := false.! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 5/25/2012 22:07'!
padWithZeros: aBoolean
	"If aBoolean is true, and totalWidth is specified, pad left with zeros, else pad with spaces"
	padWithZeros := aBoolean! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 6/2/2012 15:32'!
positiveSignChar: aCharacterOrNil
	"Set which character should be printed if number is positive (generally a + sign or a space), nil if none."
	positiveSignChar := aCharacterOrNil! !

!VersatileFloatPrinter methodsFor: 'private-printing' stamp: 'nice 6/2/2012 22:27'!
printExceptionalFloat: aFloat limitedWidthOn: aStream
	| right |
	right := String new: 32 streamContents: [:subStream | self printExceptionalFloat: aFloat on: subStream].
	self printSpaces: totalWidth - right size on: aStream.
	aStream nextPutAll: right.
	^self! !

!VersatileFloatPrinter methodsFor: 'private-printing' stamp: 'nice 9/28/2021 01:01'!
printExponent: e on: aStream
	"Print the scaling with exponent part"
	
	aStream nextPut: exponentChar.
	e >= 0 ifTrue: [forceExponentSign ifTrue: [aStream nextPut: $+]].
	exponentDigitCount > 1
		ifTrue:
			[e < 0 ifTrue: [aStream nextPut: $-].
			self printZeros: exponentDigitCount - (e numberOfDigitsInBase: 10) on: aStream.
			e abs printOn: aStream]
		ifFalse: [e printOn: aStream]! !

!VersatileFloatPrinter methodsFor: 'private-printing' stamp: 'nice 6/8/2012 21:37'!
printLeadingFractionPartZerosOn: aStream baseExpEstimate: baseExpEstimate
	"Print the leading zeros of the fraction part, before the significand digits."
	
	aStream nextPut: self fractionSeparator.
	self printZeros: 0 - baseExpEstimate on: aStream! !

!VersatileFloatPrinter methodsFor: 'private-printing' stamp: ''!
printSignOf: aFloat on: aStream
	aFloat signBit = 1
		ifTrue: [aStream nextPut: $-]
		ifFalse: [positiveSignChar ifNotNil: [aStream nextPut: positiveSignChar]]! !

!VersatileFloatPrinter methodsFor: 'private-printing' stamp: 'nice 6/2/2012 14:56'!
printSpaces: anInteger on: aStream
	"Print a given number of zeros"
	
	anInteger <= 0 ifTrue: [^self].
	aStream nextPutAll: (anInteger < 10
		ifTrue: [('         ' truncateTo: anInteger)]
		ifFalse: [String new: anInteger withAll: $ ])! !

!VersatileFloatPrinter methodsFor: 'private-printing' stamp: 'nice 6/10/2012 18:29'!
printTrailingFractionPartZerosOn: aStream decPointCount: decPointCount baseExpEstimate: baseExpEstimate
	"Print enough digits after fractionSeparator and eventually the fractionSeparator itself"
	
	| limit |
	limit := self baseExpLimit: baseExpEstimate.
	alwaysPrintFractionPart ifTrue: [limit := limit min: 0].
	(decPointCount = 1 and: [decPointCount > limit or: [alwaysPrintFractionSeparator]])
		ifTrue: [aStream nextPut: fractionSeparator].
	decPointCount > limit ifTrue: [self printZeros: decPointCount - limit on: aStream]! !

!VersatileFloatPrinter methodsFor: 'private-printing' stamp: 'nice 6/7/2012 23:13'!
printTrailingIntegerPartZerosOn: aStream decPointCount: decPointCount
	"Print the zeros at end of integer part."
	
	decPointCount > 1
		ifTrue:
			[self printZeros: decPointCount - 1 on: aStream.
			^1].
	^decPointCount! !

!VersatileFloatPrinter methodsFor: 'private-printing' stamp: 'nice 6/2/2012 14:56'!
printZeros: anInteger on: aStream
	"Print a given number of zeros"

	anInteger <= 0 ifTrue: [^self].
	aStream nextPutAll: (anInteger < 10
		ifTrue: [('000000000' truncateTo: anInteger)]
		ifFalse: [String new: anInteger withAll: $0])! !

!VersatileFloatPrinter methodsFor: 'private-measuring' stamp: 'nice 9/28/2021 01:23'!
scientificFormatSizeOf: aNumber
	"Answer the size required for printing aNumber in scientific format."

	| baseExpEstimate size integerCount fractionCount |
	baseExpEstimate := aNumber initializePrinter: self.
	
	"count sign and exponent sign"
	size := (self forceSign
				ifTrue: [1]
				ifFalse: [aNumber signBit = 1 ifTrue: [1] ifFalse: [0]])
		+ (forceExponentSign
				ifTrue: [1]
				ifFalse: [baseExpEstimate < 1 ifTrue: [1] ifFalse: [0]]).

	"Note that exponent is always printed in base 10"
	size := size + 1 + ((baseExpEstimate - 1 numberOfDigitsInBase: 10) max: exponentDigitCount).
	
	"count the digits"
	integerCount := self trueScientificIntegerCount.
	fractionCount := self trueScientificFractionCount.

	"count the fractionSeparator"
	(fractionCount > 0 or: [alwaysPrintFractionSeparator])  ifTrue: [size := size + 1].

	^size + integerCount + fractionCount! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 5/29/2012 22:23'!
showTrailingZeros: aBoolean
	showTrailingZeros := aBoolean! !

!VersatileFloatPrinter methodsFor: 'private-measuring' stamp: 'nice 9/28/2021 01:20'!
thousandFormatSizeOf: aNumber
	"Answer the size required for printing aNumber in thousand format."

	| size trueDigitCount baseExpEstimate exp expCount integerDigitCount |
	baseExpEstimate := aNumber initializePrinter: self.

	"count sign"
	size := (self forceSign or: [aNumber sign = -1]) ifTrue: [1] ifFalse: [0].

	"count the digits"
	trueDigitCount := self trueThousandDigitCount: baseExpEstimate.
		
	"count the fractionSeparator"
	integerDigitCount := baseExpEstimate - 1 \\ 3 + 1.
	(trueDigitCount > integerDigitCount or: [alwaysPrintFractionSeparator]) ifTrue: [size := size + 1].
	
	"Note that exponent is always printed in base 10"
	exp := baseExpEstimate - 1 // 3 * 3.
	expCount := exp = 0
		ifTrue: [0]
		ifFalse: [1 "exponent letter"
			+ ((exp numberOfDigitsInBase: 10) max: exponentDigitCount) "exponent digit"
			+ ((forceExponentSign or: [exp < 0]) "exponent sign"
				ifTrue: [1] ifFalse: [0])].
	
	^size + trueDigitCount + expCount! !

!VersatileFloatPrinter methodsFor: 'accessing' stamp: 'nice 5/25/2012 22:06'!
totalWidth: anInteger
	"Set the total width to be occupied by printing"
	totalWidth := anInteger! !

!VersatileFloatPrinter methodsFor: 'private-measuring' stamp: ''!
trueFixedFractionCount: baseExpEstimate
	"Answer how many digits are necessary, after the fraction separator"

	| fractionCount |
	fractionCount := digitCount isNil 
		ifTrue: [self copy countSignificandDigits - baseExpEstimate]
		ifFalse: [showTrailingZeros
			ifTrue: [1 - (self baseExpLimit: baseExpEstimate)]
			ifFalse: [digitCount + baseExpEstimate <= 0
				ifTrue: [0]
				ifFalse: [self copy countSignificandDigits - baseExpEstimate min: digitCount]]].
	fractionCount := fractionCount max: (alwaysPrintFractionPart
		ifTrue: [1]
		ifFalse: [0]).
	^fractionCount! !

!VersatileFloatPrinter methodsFor: 'private-measuring' stamp: ''!
trueFixedIntegerCount: baseExpEstimate
	"Answer how many digits are necessary before fraction"

	^baseExpEstimate max: 1! !

!VersatileFloatPrinter methodsFor: 'private-measuring' stamp: ''!
trueScientificFractionCount
	"Answer how many digits are necessary, after the fraction separator"

	| fractionCount |
	fractionCount := (digitCount notNil and: [showTrailingZeros])
		ifTrue: [digitCount]
		ifFalse: [self copy countSignificandDigits - 1].
	alwaysPrintFractionPart ifTrue: [fractionCount := fractionCount max: 1].
	^fractionCount! !

!VersatileFloatPrinter methodsFor: 'private-measuring' stamp: ''!
trueScientificIntegerCount
	"Answer how many digits are necessary, before the fraction point"

	^1! !

!VersatileFloatPrinter methodsFor: 'private-measuring' stamp: 'nice 4/7/2016 21:57'!
trueThousandDigitCount: baseExpEstimate
	"Answer how many digits are necessary, counting strictly necessary digits and optionally trailing zeros"

	| fractionCount integerCount |
	integerCount := baseExpEstimate - 1 \\ 3 + 1.
	fractionCount := (digitCount notNil and: [showTrailingZeros])
		ifTrue: [digitCount <= 0
			ifTrue: [0 - digitCount]
			ifFalse: [digitCount - integerCount]]
		ifFalse: [self countSignificandDigits - integerCount].
	fractionCount := fractionCount max: (alwaysPrintFractionPart
		ifTrue: [1]
		ifFalse: [0]).
	^integerCount + fractionCount! !

!Float methodsFor: '*NumberPrinter' stamp: 'nice 6/10/2012 21:32'!
initializePrinter: aNumberPrinter
	^aNumberPrinter setFloat: self! !

!Number methodsFor: '*NumberPrinter' stamp: 'nice 6/10/2012 21:37'!
isFinite
	^true! !

!ScaledDecimal methodsFor: '*NumberPrinter' stamp: 'nice 6/10/2012 21:33'!
initializePrinter: aNumberPrinter
	^aNumberPrinter setFraction: self asFraction! !

!Integer methodsFor: '*NumberPrinter' stamp: 'nice 6/10/2012 21:33'!
initializePrinter: aNumberPrinter
	^aNumberPrinter setFraction: self! !

!Fraction methodsFor: '*NumberPrinter' stamp: 'nice 6/10/2012 21:33'!
initializePrinter: aNumberPrinter
	^aNumberPrinter setFraction: self! !
